###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.bigquery import query_to_view
from starthinker.util.bigquery import table_create
from starthinker.util.data import get_rows
from starthinker.util.data import put_rows
from starthinker.util.google_api.discovery_to_bigquery import Discovery_To_BigQuery
from starthinker.util.project import project
from starthinker.util.regexp import lookup_id
from starthinker.util.sheets import sheets_clear

from starthinker.task.dv_editor.insertion_order import insertion_order_commit
from starthinker.task.dv_editor.line_item import line_item_commit
from starthinker.task.dv_editor.advertiser import advertiser_commit
from starthinker.task.dv_editor.patch import patch_masks
from starthinker.task.dv_editor.patch import patch_preview


def integration_detail_clear():
  sheets_clear(project.task["auth_sheets"], project.task["sheet"],
               "Integration Details", "A2:Z")


def integration_detail_load():

  # write integration_details to sheet
  rows = get_rows(
      project.task["auth_bigquery"], {
          "bigquery": {
              "dataset":
                  project.task["dataset"],
              "query":
                  """SELECT
         CONCAT(P.displayName, ' - ', P.partnerId),
         CONCAT(A.displayName, ' - ', A.advertiserId),
         NULL,
         NULL,
         NULL,
         A.integrationDetails.integrationCode,
         A.integrationDetails.integrationCode,
         A.integrationDetails.details,
         A.integrationDetails.details
       FROM `{dataset}.DV_Advertisers` AS A
       LEFT JOIN `{dataset}.DV_Partners` AS P
       ON A.partnerId=P.partnerId
       UNION ALL
       SELECT
         CONCAT(P.displayName, ' - ', P.partnerId),
         CONCAT(A.displayName, ' - ', A.advertiserId),
         CONCAT(C.displayName, ' - ', C.campaignId),
         CONCAT(I.displayName, ' - ', I.insertionOrderId),
         NULL,
         I.integrationDetails.integrationCode,
         I.integrationDetails.integrationCode,
         I.integrationDetails.details,
         I.integrationDetails.details
       FROM `{dataset}.DV_InsertionOrders` AS I
       LEFT JOIN `{dataset}.DV_Campaigns` AS C
       ON I.campaignId=C.campaignId
       LEFT JOIN `{dataset}.DV_Advertisers` AS A
       ON I.advertiserId=A.advertiserId
       LEFT JOIN `{dataset}.DV_Partners` AS P
       ON A.partnerId=P.partnerId
       UNION ALL
       SELECT
         CONCAT(P.displayName, ' - ', P.partnerId),
         CONCAT(A.displayName, ' - ', A.advertiserId),
         CONCAT(C.displayName, ' - ', C.campaignId),
         CONCAT(I.displayName, ' - ', I.insertionOrderId),
         CONCAT(L.displayName, ' - ', L.lineItemId),
         L.integrationDetails.integrationCode,
         L.integrationDetails.integrationCode,
         L.integrationDetails.details,
         L.integrationDetails.details
       FROM `{dataset}.DV_LineItems` AS L
       LEFT JOIN `{dataset}.DV_Campaigns` AS C
       ON L.campaignId=C.campaignId
       LEFT JOIN `{dataset}.DV_InsertionOrders` AS I
       ON L.insertionOrderId=I.insertionOrderId
       LEFT JOIN `{dataset}.DV_Advertisers` AS A
       ON L.advertiserId=A.advertiserId
       LEFT JOIN `{dataset}.DV_Partners` AS P
       ON A.partnerId=P.partnerId
       """.format(**project.task),
              "legacy":
                  False
          }
      })

  put_rows(
      project.task["auth_sheets"], {
          "sheets": {
              "sheet": project.task["sheet"],
              "tab": "Integration Details",
              "range": "A2"
          }
      }, rows)


def integration_detail_audit():
  rows = get_rows(
      project.task["auth_sheets"], {
          "sheets": {
              "sheet": project.task["sheet"],
              "tab": "Integration Details",
              "range": "A2:Z"
          }
      })

  put_rows(
      project.task["auth_bigquery"], {
          "bigquery": {
              "dataset": project.task["dataset"],
              "table": "SHEET_IntegrationDetails",
              "schema": [
                  { "name": "Partner", "type": "STRING" },
                  { "name": "Advertiser", "type": "STRING" },
                  { "name": "Campaign", "type": "STRING" },
                  { "name": "Insertion_Order", "type": "STRING" },
                  { "name": "Line_Item", "type": "STRING" },
                  { "name": "Integration_Code", "type": "STRING" },
                  { "name": "Integration_Code_Edit", "type": "STRING" },
                  { "name": "Details", "type": "STRING" },
                  { "name": "Details_Edit", "type": "STRING" },
              ],
              "format": "CSV"
          }
      }, rows)

  query_to_view(
    project.task["auth_bigquery"],
    project.id,
    project.task["dataset"],
    "AUDIT_IntegrationDetails",
    """WITH
      /* Check if advertiser values are set */
      INPUT_ERRORS AS (
        SELECT
          'Integration Details' AS Operation,
          'Missing Advertiser.' AS Error,
          'ERROR' AS Severity,
          COALESCE(Line_Item, Insertion_Order, Advertiser, 'BLANK') AS Id
        FROM
          `{dataset}.SHEET_IntegrationDetails`
        WHERE Advertiser IS NULL
      )

      SELECT * FROM INPUT_ERRORS
    """.format(**project.task),
    legacy=False
  )

  query_to_view(
    project.task["auth_bigquery"],
    project.id,
    project.task["dataset"],
    "PATCH_IntegrationDetails",
    """SELECT *
      FROM `{dataset}.SHEET_IntegrationDetails`
      WHERE Line_Item NOT IN (SELECT Id FROM `{dataset}.AUDIT_IntegrationDetails` WHERE Severity='ERROR')
      AND Insertion_Order NOT IN (SELECT Id FROM `{dataset}.AUDIT_IntegrationDetails` WHERE Severity='ERROR')
      AND Campaign NOT IN (SELECT Id FROM `{dataset}.AUDIT_IntegrationDetails` WHERE Severity='ERROR')
    """.format(**project.task),
    legacy=False
  )


def integration_detail_patch(commit=False):
  patches = []

  rows = get_rows(
    project.task["auth_bigquery"],
    { "bigquery": {
      "dataset": project.task["dataset"],
      "table":"PATCH_IntegrationDetails",
    }},
    as_object=True
  )

  for row in rows:

    integration_details = {}

    if row['Integration_Code'] != row['Integration_Code_Edit']:
      integration_details.setdefault("integrationDetails", {})
      integration_details["integrationDetails"]["integrationCode"] = row['Integration_Code_Edit']
    if row['Details'] != row['Details_Edit']:
      integration_details.setdefault("integrationDetails", {})
      integration_details["integrationDetails"]["details"] = row['Details_Edit']

    if integration_details:
      patch = {
          "operation": "Pacing",
          "action": "PATCH",
          "partner": row['Partner'],
          "parameters": {
              "advertiserId": lookup_id(row['Advertiser']),
              "body": integration_details
          }
      }

      if row['Line_Item']:
        patch["line_item"] = row['Line_Item']
        patch["parameters"]["lineItemId"] = lookup_id(row['Line_Item'])

      elif row['Insertion_Order']:
        patch["insertion_order"] = row['Insertion_Order']
        patch["parameters"]["insertionOrderId"] = lookup_id(row['Insertion_Order'])

      else:
        patch["advertiser"] = row['Advertiser']
        patch["parameters"]["advertiserId"] = lookup_id(row['Advertiser'])

      patches.append(patch)

  patch_masks(patches)

  if commit:
    insertion_order_commit(patches)
    line_item_commit(patches)
    advertiser_commit(patches)
  else:
    patch_preview(patches)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer ( recommended ):

    From Release: pip install starthinker
    From Open Source: pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory ( if pushing local code changes ):

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

  If any recipe task has "auth" set to "user" add user credentials:

    1. Ensure an RECIPE['setup']['auth']['user'] = [User Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_user", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/deploy_commandline.md#optional-setup-user-credentials

--------------------------------------------------------------

  If any recipe task has "auth" set to "service" add service credentials:

    1. Ensure an RECIPE['setup']['auth']['service'] = [Service Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_service", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md

--------------------------------------------------------------

cTV Inventory Availability Dashboard

The cTV Audience Affinity dashboard is designed to give clients insights into which cTV apps their audiences have a high affinity for using.  The goal of this dashboard is to provide some assistance with the lack of audience targeting for cTV within DV360.

  - Find instructions and recommendations for this dashboard <a href="https://docs.google.com/document/d/120kcR9OrS4hGdTxRK0Ig2koNmm6Gl7sH0L6U56N0SAM/view?usp=sharing" target="_blank">here</a>

--------------------------------------------------------------

This StarThinker DAG can be extended with any additional tasks from the following sources:
  - https://google.github.io/starthinker/
  - https://github.com/google/starthinker/tree/master/dags

'''

from starthinker.airflow.factory import DAG_Factory

INPUTS = {
  'dataset': '',  # BigQuery Dataset where all data will live.
  'recipe_project': '',  # Project where BigQuery dataset will be created.
  'recipe_name': '',  # Name of document to deploy to.
  'auth_write': 'service',  # Credentials used for writing data.
  'partner_id': '',  # DV360 Partner id.
  'auth_read': 'user',  # Credentials used for reading data.
  'audience_ids': '',  # Comma separated list of Audience Ids
}

RECIPE = {
  'tasks': [
    {
      'drive': {
        'auth': 'user',
        'copy': {
          'source': 'https://docs.google.com/spreadsheets/d/1PPPk2b4gGJHNgQ4hXLiTKzH8pRIdlF5fNy9VCw1v7tM/',
          'destination': {
            'field': {
              'name': 'recipe_name',
              'prefix': 'cTV App Match Table ',
              'kind': 'string',
              'order': 1,
              'description': 'Name of document to deploy to.',
              'default': ''
            }
          }
        }
      }
    },
    {
      'dataset': {
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'dataset': {
          'field': {
            'name': 'dataset',
            'kind': 'string',
            'order': 3,
            'default': '',
            'description': 'BigQuery Dataset where all data will live.'
          }
        }
      }
    },
    {
      'dbm': {
        'auth': 'user',
        'report': {
          'body': {
            'timezoneCode': 'America/Los_Angeles',
            'kind': 'doubleclickbidmanager#query',
            'metadata': {
              'title': {
                'field': {
                  'name': 'recipe_name',
                  'kind': 'string',
                  'prefix': 'us_country_app_'
                }
              },
              'dataRange': 'LAST_30_DAYS',
              'format': 'CSV',
              'sendNotification': False
            },
            'params': {
              'type': 'TYPE_INVENTORY_AVAILABILITY',
              'groupBys': [
                'FILTER_APP_URL'
              ],
              'filters': [
                {
                  'type': 'FILTER_PARTNER',
                  'value': {
                    'field': {
                      'name': 'partner_id',
                      'kind': 'integer',
                      'order': 1,
                      'description': 'DV360 Partner id.'
                    }
                  }
                },
                {
                  'type': 'FILTER_INVENTORY_FORMAT',
                  'value': 'VIDEO'
                },
                {
                  'type': 'FILTER_COUNTRY',
                  'value': 'US'
                }
              ],
              'metrics': [
                'METRIC_BID_REQUESTS',
                'METRIC_UNIQUE_VISITORS_COOKIES'
              ],
              'includeInviteData': True
            },
            'schedule': {
              'frequency': 'DAILY',
              'nextRunMinuteOfDay': 0,
              'nextRunTimezoneCode': 'America/Los_Angeles',
              'endTimeMs': 7983727200000
            }
          }
        },
        'out': {
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'dataset',
                'kind': 'string',
                'order': 3,
                'default': '',
                'description': 'BigQuery Dataset where all data will live.'
              }
            },
            'schema': [
              {
                'type': 'STRING',
                'name': 'app_url',
                'mode': 'NULLABLE'
              },
              {
                'type': 'STRING',
                'name': 'impressions',
                'mode': 'NULLABLE'
              },
              {
                'type': 'STRING',
                'name': 'uniques',
                'mode': 'NULLABLE'
              }
            ],
            'table': 'us_country_app'
          }
        }
      }
    },
    {
      'dbm': {
        'auth': 'user',
        'report': {
          'body': {
            'timezoneCode': 'America/Los_Angeles',
            'kind': 'doubleclickbidmanager#query',
            'metadata': {
              'title': {
                'field': {
                  'name': 'recipe_name',
                  'kind': 'string',
                  'prefix': 'us_country_baseline_'
                }
              },
              'dataRange': 'LAST_30_DAYS',
              'format': 'CSV',
              'sendNotification': False
            },
            'params': {
              'type': 'TYPE_INVENTORY_AVAILABILITY',
              'filters': [
                {
                  'type': 'FILTER_PARTNER',
                  'value': {
                    'field': {
                      'name': 'partner_id',
                      'kind': 'integer',
                      'order': 1,
                      'description': 'DV360 Partner id.'
                    }
                  }
                },
                {
                  'type': 'FILTER_COUNTRY',
                  'value': 'US'
                }
              ],
              'metrics': [
                'METRIC_BID_REQUESTS',
                'METRIC_UNIQUE_VISITORS_COOKIES'
              ],
              'includeInviteData': True
            },
            'schedule': {
              'frequency': 'DAILY',
              'nextRunMinuteOfDay': 0,
              'nextRunTimezoneCode': 'America/Los_Angeles',
              'endTimeMs': 7983727200000
            }
          }
        },
        'out': {
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'dataset',
                'kind': 'string',
                'order': 3,
                'default': '',
                'description': 'BigQuery Dataset where all data will live.'
              }
            },
            'schema': [
              {
                'type': 'STRING',
                'name': 'impressions',
                'mode': 'NULLABLE'
              },
              {
                'type': 'STRING',
                'name': 'uniques',
                'mode': 'NULLABLE'
              }
            ],
            'table': 'us_country_baseline'
          }
        }
      }
    },
    {
      'dbm': {
        'auth': 'user',
        'report': {
          'filters': {
            'FILTER_USER_LIST': {
              'single_cell': True,
              'values': {
                'field': {
                  'name': 'audience_ids',
                  'kind': 'integer_list',
                  'order': 2,
                  'description': 'Comma separated list of Audience Ids'
                }
              }
            }
          },
          'body': {
            'timezoneCode': 'America/Los_Angeles',
            'kind': 'doubleclickbidmanager#query',
            'metadata': {
              'title': {
                'field': {
                  'name': 'recipe_name',
                  'kind': 'string',
                  'prefix': 'us_audience_baseline_'
                }
              },
              'dataRange': 'LAST_30_DAYS',
              'format': 'CSV',
              'sendNotification': False
            },
            'params': {
              'type': 'TYPE_INVENTORY_AVAILABILITY',
              'groupBys': [
                'FILTER_AUDIENCE_LIST'
              ],
              'filters': [
                {
                  'type': 'FILTER_PARTNER',
                  'value': {
                    'field': {
                      'name': 'partner_id',
                      'kind': 'integer',
                      'order': 1,
                      'description': 'DV360 Partner id.'
                    }
                  }
                },
                {
                  'type': 'FILTER_COUNTRY',
                  'value': 'US'
                }
              ],
              'metrics': [
                'METRIC_BID_REQUESTS',
                'METRIC_UNIQUE_VISITORS_COOKIES'
              ],
              'includeInviteData': True
            },
            'schedule': {
              'frequency': 'DAILY',
              'nextRunMinuteOfDay': 0,
              'nextRunTimezoneCode': 'America/Los_Angeles',
              'endTimeMs': 7983727200000
            }
          }
        },
        'out': {
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'dataset',
                'kind': 'string',
                'order': 3,
                'default': '',
                'description': 'BigQuery Dataset where all data will live.'
              }
            },
            'schema': [
              {
                'type': 'STRING',
                'name': 'user_list',
                'mode': 'NULLABLE'
              },
              {
                'type': 'STRING',
                'name': 'impressions',
                'mode': 'NULLABLE'
              },
              {
                'type': 'STRING',
                'name': 'uniques',
                'mode': 'NULLABLE'
              }
            ],
            'table': 'us_audience_baseline'
          }
        }
      }
    },
    {
      'dbm': {
        'auth': 'user',
        'report': {
          'filters': {
            'FILTER_USER_LIST': {
              'single_cell': True,
              'values': {
                'field': {
                  'name': 'audience_ids',
                  'kind': 'integer_list',
                  'order': 2,
                  'description': 'Comma separated list of Audience Ids'
                }
              }
            }
          },
          'body': {
            'timezoneCode': 'America/Los_Angeles',
            'kind': 'doubleclickbidmanager#query',
            'metadata': {
              'title': {
                'field': {
                  'name': 'recipe_name',
                  'kind': 'string',
                  'prefix': 'us_audience_app_'
                }
              },
              'dataRange': 'LAST_30_DAYS',
              'format': 'CSV',
              'sendNotification': False
            },
            'params': {
              'type': 'TYPE_INVENTORY_AVAILABILITY',
              'groupBys': [
                'FILTER_APP_URL',
                'FILTER_AUDIENCE_LIST'
              ],
              'filters': [
                {
                  'type': 'FILTER_PARTNER',
                  'value': {
                    'field': {
                      'name': 'partner_id',
                      'kind': 'integer',
                      'order': 1,
                      'description': 'DV360 Partner id.'
                    }
                  }
                },
                {
                  'type': 'FILTER_INVENTORY_FORMAT',
                  'value': 'VIDEO'
                },
                {
                  'type': 'FILTER_COUNTRY',
                  'value': 'US'
                }
              ],
              'metrics': [
                'METRIC_BID_REQUESTS',
                'METRIC_UNIQUE_VISITORS_COOKIES'
              ],
              'includeInviteData': True
            },
            'schedule': {
              'frequency': 'DAILY',
              'nextRunMinuteOfDay': 0,
              'nextRunTimezoneCode': 'America/Los_Angeles',
              'endTimeMs': 7983727200000
            }
          }
        },
        'out': {
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'dataset',
                'kind': 'string',
                'order': 3,
                'default': '',
                'description': 'BigQuery Dataset where all data will live.'
              }
            },
            'schema': [
              {
                'type': 'STRING',
                'name': 'app_url',
                'mode': 'NULLABLE'
              },
              {
                'type': 'STRING',
                'name': 'user_list',
                'mode': 'NULLABLE'
              },
              {
                'type': 'STRING',
                'name': 'impressions',
                'mode': 'NULLABLE'
              },
              {
                'type': 'STRING',
                'name': 'uniques',
                'mode': 'NULLABLE'
              }
            ],
            'table': 'us_audience_app'
          }
        }
      }
    },
    {
      'sheets': {
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 1,
            'default': 'user',
            'description': 'Credentials used for reading data.'
          }
        },
        'sheet': {
          'field': {
            'name': 'recipe_name',
            'prefix': 'cTV App Match Table ',
            'kind': 'string',
            'order': 1,
            'description': 'Name of document to deploy to.',
            'default': ''
          }
        },
        'tab': 'data',
        'range': 'A:Z',
        'header': True,
        'out': {
          'auth': {
            'field': {
              'name': 'auth_write',
              'kind': 'authentication',
              'order': 1,
              'default': 'service',
              'description': 'Credentials used for writing data.'
            }
          },
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'dataset',
                'kind': 'string',
                'description': 'BigQuery Dataset where all data will live.'
              }
            },
            'table': 'CTV_App_Lookup',
            'schema': [
              {
                'type': 'STRING',
                'name': 'Publisher_Name',
                'mode': 'NULLABLE'
              },
              {
                'type': 'STRING',
                'name': 'CTV_App_name',
                'mode': 'NULLABLE'
              }
            ]
          }
        }
      }
    },
    {
      'bigquery': {
        'description': 'The query to join all the IAR reports into an Affinity Index.',
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'from': {
          'query': "SELECT    audience_app.app_url,    audience_app.ctv_app_name,  IF    (audience_app.app_url LIKE '%Android%'      OR audience_app.app_url LIKE '%iOS',      'App',      'Domain') AS app_or_domain,    audience_app.user_list AS audience_list,    audience_app.Potential_Impressions AS audience_app_impressions,    audience_app.Unique_Cookies_With_Impressions AS audience_app_uniques,    audience_baseline.Potential_Impressions AS audience_baseline_impressions,    audience_baseline.Unique_Cookies_With_Impressions AS audience_baseline_uniques,    country_app.Potential_Impressions AS country_app_impressions,    country_app.Unique_Cookies_With_Impressions AS country_app_uniques,    country_baseline.Potential_Impressions AS country_baseline_impressions,    country_baseline.Unique_Cookies_With_Impressions AS country_baseline_uniques,    ((audience_app.Unique_Cookies_With_Impressions/NULLIF(audience_baseline.Unique_Cookies_With_Impressions,          0))/NULLIF((country_app.Unique_Cookies_With_Impressions/NULLIF(CAST(country_baseline.Unique_Cookies_With_Impressions AS int64),            0)),        0))*100 AS affinity_index  FROM (    SELECT      user_list,      CAST(      IF        (impressions LIKE '%< 1000%',          0,          CAST(impressions AS int64)) AS int64) AS potential_impressions,      CAST(      IF        (uniques LIKE '%< 100%',          0,          CAST(uniques AS int64)) AS int64) AS unique_cookies_with_impressions    FROM      `[PARAMETER].[PARAMETER].us_audience_baseline` ) AS audience_baseline  JOIN (    SELECT      ctv_app.CTV_App_name AS ctv_app_name,      user_list,      app_url,      CAST(      IF        (impressions LIKE '%< 1000%',          0,          CAST(impressions AS int64)) AS int64) AS potential_impressions,      CAST(      IF        (uniques LIKE '%< 1000%',          0,          CAST(uniques AS int64)) AS int64) AS unique_cookies_with_impressions    FROM      `[PARAMETER].[PARAMETER].us_audience_app` AS a    LEFT JOIN      `[PARAMETER].[PARAMETER].CTV_App_Lookup` AS ctv_app    ON      a.app_url = ctv_app.Publisher_Name ) AS audience_app  ON    audience_baseline.user_list = audience_app.user_list  LEFT JOIN (    SELECT      app_url,      CAST(      IF        (CAST(impressions AS STRING) LIKE '%< 1000%',          0,          CAST(impressions AS int64)) AS int64) AS Potential_Impressions,      CAST(      IF        (CAST(uniques AS STRING) LIKE '%< 1000%',          0,          CAST(uniques AS int64)) AS int64) AS Unique_Cookies_With_Impressions    FROM      `[PARAMETER].[PARAMETER].us_country_app` ) AS country_app  ON    country_app.app_url = audience_app.app_url  CROSS JOIN (    SELECT      CAST(      IF        (CAST(impressions AS STRING) LIKE '%< 1000%',          0,          CAST(impressions AS int64)) AS int64) AS Potential_Impressions,      CAST(      IF        (CAST(uniques AS STRING) LIKE '%< 1000%',          0,          CAST(uniques AS int64)) AS int64) AS Unique_Cookies_With_Impressions    FROM      `[PARAMETER].[PARAMETER].us_country_baseline` ) AS country_baseline",
          'legacy': False,
          'parameters': [
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'dataset',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'dataset',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'dataset',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'dataset',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'dataset',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'dataset',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            }
          ]
        },
        'to': {
          'dataset': {
            'field': {
              'name': 'dataset',
              'kind': 'string',
              'description': 'BigQuery Dataset where all data will live.'
            }
          },
          'table': 'final_table'
        }
      }
    }
  ]
}

DAG_FACTORY = DAG_Factory('ctv_audience_affinity', RECIPE, INPUTS)
DAG = DAG_FACTORY.generate()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import json
from functools import wraps

from django.conf import settings
from django.http import HttpResponseRedirect

from starthinker.util.auth.wrapper import CredentialsFlowWrapper


def permission_admin():

  def _decorator(_view):

    @wraps(_view)
    def _wrapper(request, *args, **kwargs):

      # user is logged in
      if request.user.is_authenticated:
        return _view(request, *args, **kwargs)

      # multi user mode, log user in using oauth
      else:
        flow = CredentialsFlowWrapper(
            settings.UI_CLIENT,
            redirect_uri=settings.CONST_URL + '/oauth_callback/')
        auth_url, _ = flow.authorization_url(
            prompt='', access_type='offline', include_granted_scopes='true')
        request.session['code_verifier'] = flow.code_verifier
        return HttpResponseRedirect(auth_url)

    return _wrapper

  return _decorator



# Generated by Django 2.2.13 on 2020-11-19 10:59

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Project',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('identifier', models.CharField(max_length=255)),
                ('service', models.TextField()),
                ('share', models.CharField(default='', max_length=50)),
                ('account', models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from django.conf.urls import url

from starthinker_ui.recipe import views

urlpatterns = [
    url(r'^recipe/edit/(?P<pk>\d+)?/?$', views.recipe_edit, name='recipe.edit'),
    url(r'^recipe/manual/(?P<pk>\d+)?/?$',
        views.recipe_manual,
        name='recipe.manual'),
    url(r'^recipe/delete/(?P<pk>\d+)/$',
        views.recipe_delete,
        name='recipe.delete'),
    url(r'^recipe/run/(?P<pk>\d+)/$', views.recipe_run, name='recipe.run'),
    url(r'^recipe/cancel/(?P<pk>\d+)/$',
        views.recipe_cancel,
        name='recipe.cancel'),
    url(r'^recipe/status/(?P<pk>\d+)/$',
        views.recipe_status,
        name='recipe.status'),
    url(r'^recipe/download/(?P<pk>\d+)?/?$',
        views.recipe_download,
        name='recipe.download'),
    url(r'^recipe/download/json/(?P<pk>\d+)/$',
        views.recipe_json,
        name='recipe.json'),
    url(r'^recipe/download/colab/(?P<pk>\d+)/$',
        views.recipe_colab,
        name='recipe.colab'),
    url(r'^recipe/download/airflow/(?P<pk>\d+)/$',
        views.recipe_airflow,
        name='recipe.airflow'),
    url(r'^recipe/start/$', views.recipe_start, name='recipe.start'),
    url(r'^recipe/stop/$', views.recipe_stop, name='recipe.stop'),
    url(r'^$', views.recipe_list, name='recipe.list'),
    url(r'^recipe/autoscale/$', views.autoscale, name='recipe.autoscale'),
]



import os
from itertools import chain

from django.core.management.base import BaseCommand, CommandError
from django.template.loader import render_to_string
from django.conf import settings

from starthinker_ui.recipe.scripts import Script


class Command(BaseCommand):
  help = 'Generate GTECH Files For Google3'

  def handle(self, *args, **kwargs):

    # hard code to be the internal server
    settings.CONST_URL = "https://starthinker.corp.google.com"

    scripts = list(Script.get_scripts())

    print('Writing:', '%s/docs/GTECH' % settings.UI_ROOT)

    with open('%s/docs/GTECH' % settings.UI_ROOT, 'w') as gtech_file:
      gtech_file.write(
          render_to_string(
              'website/starthinker.gtech', {
                  'scripts':
                      len(scripts),
                  'authors':
                      len(
                          set(
                              chain(
                                  *[script.get_authors()
                                    for script in scripts])))
              }))

    scripts = [
        s for s in Script.get_scripts()
        if s.get_impacts() and s.get_open_source()
    ]

    for script in scripts:

      directory = '%s/docs/solution/%s' % (settings.UI_ROOT, script.get_tag())
      print('Writing: %s/GTECH' % directory)
      if not os.path.exists(directory):
        os.makedirs(directory)

      with open('%s/GTECH' % directory, 'w') as solution_file:
        solution_file.write(
            render_to_string('website/script.gtech', {'script': script}))



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

Floodlight_Schema = [{
    'name': 'Partitiontime',
    'type': 'TIMESTAMP',
    'mode': 'NULLABLE'
}, {
    'name': 'Event_Time',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'User_Id',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Advertiser_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Campaign_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Ad_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Version',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Site_Id_Dcm',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Browser_Platform_Version',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Event_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Event_Sub_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_Auction_Id',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_Request_Time',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_Advertiser_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_Line_Item_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_Creative_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_Url',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_Site_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_Language',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_Adx_Page_Categories',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_Matching_Targeted_Keywords',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_Exchange_Id',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_Attributed_Inventory_Source_External_Id',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_Attributed_Inventory_Source_Is_Public',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_Ad_Position',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_Country_Code',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_Browser_Platform_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_Net_Speed',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_Matching_Targeted_Segments',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_Isp_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_Mobile_Model_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_Designated_Market_Area_Dma_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_City_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_Browser_Timezone_Offset_Minutes',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_Mobile_Make_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Activity_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Pixel_Size',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Total_Conversions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}]



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Script that executes { "entity":{...}} task.

This script translates JSON instructions into operations on Entity Read File
transfer to BigQuery.

### CAUTION

ONLY PARTNER LEVEL LOG FILES ARE MOVED, no advertiser filters in
this script. To filter by advertiser, add queries within BigQuery to create
additional tables.

See: https://developers.google.com/bid-manager/guides/entity-read/format-v2

### Table Names And Entities

The prefix value specified in JSON is used to name each table and avoid
collisions.  All tables will be prefixed with "Entity_" by default. For
example 'Campaign' Entity Read becomes, 'Entity_Campaign'. You can
modify the preifx in the JSON script.

The following Entity Read Files Can Be Specified:

- Campaign
- LineItem
- Creative
- UserList
- Partner
- Advertiser
- InsertionOrder
- Pixel
- InventorySource
- CustomAffinity
- UniversalChannel
- UniversalSite
- SupportedExchange
- DataPartner
- GeoLocation
- Language
- DeviceCriteria
- Browser
- Isp

### Notes

- Files are moved using an in memory buffer, controlled by BUFFER_SCALE in
config.py.
- Bigger buffer means faster move but needs a machine with more memory.
- These transfers take a long time, two jobs can write over each other.
- The tables are filled over time, if you need instant data switch, use table
copy after job.

"""

from starthinker.config import BUFFER_SCALE
from starthinker.util.project import project
from starthinker.util.storage import object_get_chunks
from starthinker.util.bigquery import json_to_table
from starthinker.util.data import get_rows
from starthinker.task.entity.schema import Entity_Schema_Lookup

CHUNK_SIZE = int(200 * 1024000 * BUFFER_SCALE)  # 200 MB * scale in config.py
PUBLIC_FILES = ('SupportedExchange', 'DataPartner', 'UniversalSite',
                'GeoLocation', 'Language', 'DeviceCriteria', 'Browser', 'Isp')


def get_entity(path):
  delimiter = ',\r\r'
  first = True
  view = ''

  for chunk in object_get_chunks(project.task['auth'], path, CHUNK_SIZE):
    # read the next chunk, remove all newlines, leaving only '\r\r' between records ( clever use of non display characters for parsing )
    view += chunk.decode().replace('\n', '')

    # first time through, scrap the leading bracket
    if first:
      view = view.strip('[\r\r')
      first = False

    # after replacing all newlines, only '\r\r' are left, clever Googlers
    end = view.rfind(delimiter)
    if end > -1:
      yield view[:end].replace(delimiter, '\n')
      view = view[end + 1:]

  # last one never delimits, so opportunity to trim extra bracket
  yield view.strip('\r\r]')


def move_entity(project, path, table, schema, disposition):
  if 'prefix' in project.task:
    table = '%s_%s' % (project.task['prefix'], table)

  if 'out' in project.task:
    auth = project.task['out']['bigquery'].get('auth', project.task['auth'])
    dataset = project.task['out']['bigquery']['dataset']
  # deprecated, need out to allow auth mix on in and out
  else:
    auth = project.task['auth']
    dataset = project.task['dataset']

  # read the entity file in parts
  records = get_entity(path)

  # write each part
  json_to_table(
      auth,
      project.id,
      dataset,
      table,
      records,
      schema=schema,
      disposition=disposition)
  # after first write switch to append
  #disposition = 'WRITE_APPEND'


@project.from_parameters
def entity():
  if project.verbose:
    print('ENTITY')

  # legacy translations ( changed partners, advertisers to accounts with "partner_id:advertiser_id" )
  if 'partner_id' in project.task:
    project.task['accounts'] = [project.task['partner_id']]

  # create entities
  for entity in project.task['entities']:
    if project.verbose:
      print('ENTITY:', entity)

    # write public files only once
    if entity in PUBLIC_FILES:
      path = 'gdbm-public:entity/%s.0.%s.json' % (
          project.date.strftime('%Y%m%d'), entity)
      schema = Entity_Schema_Lookup[entity]
      move_entity(project, path, entity, schema, 'WRITE_TRUNCATE')

    # supports multiple partners, first one resets table, others append
    else:
      disposition = 'WRITE_TRUNCATE'
      for account in get_rows('user', project.task['partners']):

        #for account in project.task['accounts']:
        # if advertiser given do not run it ( SAFETY )
        if ':' in str(account):
          print('WARNING: Skipping advertiser: ', account)
          continue
        if project.verbose:
          print('PARTNER:', account)
        path = 'gdbm-%s:entity/%s.0.%s.json' % (
            account, project.date.strftime('%Y%m%d'), entity)
        schema = Entity_Schema_Lookup[entity]
        move_entity(project, path, entity, schema, disposition)
        disposition = 'WRITE_APPEND'


if __name__ == '__main__':
  entity()



# Generated by Django 2.2.13 on 2020-11-20 07:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('project', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='project',
            name='key',
            field=models.CharField(default='', max_length=50),
        ),
    ]



# -*- coding: utf-8 -*-

###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import json

from django.db import models


class JsonField(models.TextField):

  # python to SQL
  def get_prep_value(self, value):
    if value is None:
      return None
    elif isinstance(value, str):
      return value
    else:
      try:
        return json.dumps(value)
      except:        raise ValidationError(_('Bad JSON string.'))

  # SQL to python
  def from_db_value(self, value, expression, connection, context):
    return json.loads(value)

  # any source to python
  def to_python(self, value):
    if value is None:
      return None
    if isinstance(value, str):
      try:
        return json.loads(value)
      except:        raise ValidationError(_('Bad JSON string.'))
    else:
      return value



# Generated by Django 2.2.13 on 2020-11-19 10:59

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import starthinker_ui.recipe.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('project', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Recipe',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('token', models.CharField(max_length=8, unique=True)),
                ('reference', models.CharField(default=starthinker_ui.recipe.models.reference_default, max_length=32, unique=True)),
                ('name', models.CharField(max_length=64)),
                ('active', models.BooleanField(default=True)),
                ('manual', models.BooleanField(default=False)),
                ('week', models.CharField(default='["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"]', max_length=64)),
                ('hour', models.CharField(default='[3]', max_length=128)),
                ('timezone', models.CharField(blank=True, default='America/Los_Angeles', max_length=32)),
                ('tasks', models.TextField()),
                ('job_utm', models.BigIntegerField(blank=True, default=0)),
                ('job_status', models.TextField(default='{}')),
                ('worker_uid', models.CharField(default='', max_length=128)),
                ('worker_utm', models.BigIntegerField(blank=True, default=0)),
                ('birthday', models.DateField(auto_now_add=True)),
                ('account', models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
                ('project', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='project.Project')),
            ],
        ),
    ]



#!/usr/bin/env python

###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

# -*- coding: utf-8 -*-

from setuptools import find_packages
from setuptools import setup

REQUIREMENTS = [
    'google-api-python-client',
    'google-auth',
    'google-auth-oauthlib',
    'google-auth-httplib2',
    'google-cloud-bigquery',
    'pysftp',
    'pytz',
    'tzlocal',
    'TwitterAPI',
    'python-dateutil',
    'pandas',
    'psutil',
    'moviepy==1.0.0',
    'py-xlsx',
    'simple-salesforce',
    'smartsheet-python-sdk'
]

setup(
    name='starthinker',
    version='1.0.10',
    description='StarThinker is a Google gTech built python framework for creating and sharing re-usable workflow components.',
    long_description='StarThinker is a Google gTech built python framework for creating and sharing re-usable workflow components. To make it easier for partners and clients to work with some of our advertsing solutions, the gTech team has open sourced this framework as a reference implementation.  Our goal is to make managing data workflows using Google Cloud as fast and re-usable as possible, allowing teams to focus on building advertising solutions.',
    author='Paul Kenjora',
    author_email='kenjora@google.com',
    url='https://github.com/google/starthinker',
    packages=find_packages(),
    package_dir={'starthinker': 'starthinker'},
    include_package_data=True,
    install_requires=REQUIREMENTS,
    entry_points={
        'console_scripts': [
            'st_auth = starthinker.auth.helper:main',
            'st_json = starthinker.script.helper:main',
            'st_script = starthinker.script.run:main',
            'st_run = starthinker.all.helper:main',
            'st_cm = starthinker.task.dcm.helper:main',
            'st_dv360 = starthinker.task.dbm.helper:main',
            'st_dv360_lineitem = starthinker.task.dv360_beta.helper:main',
            'st_dv360_lineitem_legacy = starthinker.task.lineitem.helper:main',
            'st_bigquery = starthinker.task.bigquery.helper:main',
            'st_google_api = starthinker.task.google_api.helper:main',
            'st_newsletter = starthinker.task.newsletter.helper:main'
        ]
    },
    license='Apache License, Version 2.0',
    zip_safe=False,
    keywords='starthinker',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
    ]
)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Handles interactions with the Buldozer feed.

This module has all the column name mappings, reads and writes data to the
Google Sheet that represents the Bulkdozer feed, and
"""

import json
import pytz

from starthinker.util.sheets import sheets_get, sheets_read, sheets_write
from dateutil import parser


class FieldMap:
  """Contains static maps of column names and enumerations.

  This class maps user friendly field names with API enumeration values, it also
  maps Bulkdozer feed column names to constants so those can be referenced
  indirectly.
  """
  THIRD_PARTY_URL_TYPE_MAP = {
      'Start': 'VIDEO_START',
      'First quartile': 'VIDEO_FIRST_QUARTILE',
      'Midpoint': 'VIDEO_MIDPOINT',
      'Third quartile': 'VIDEO_THIRD_QUARTILE',
      'Complete': 'VIDEO_COMPLETE',
      'Mute': 'VIDEO_MUTE',
      'Pause': 'VIDEO_PAUSE',
      'Rewind': 'VIDEO_REWIND',
      'Full screen': 'VIDEO_FULLSCREEN',
      'Stop': 'VIDEO_STOP',
      'Custom click': 'CLICK_TRACKING',
      'Skip': 'VIDEO_SKIP',
      'Progress': 'VIDEO_PROGRESS'
  }

  AD_ID = 'Ad ID'
  AD_NAME = 'Ad Name'
  AD_START_DATE = 'Ad Start Date'
  AD_END_DATE = 'Ad End Date'
  AD_ACTIVE = 'Ad Active'
  AD_ARCHIVED = 'Ad Archived'
  AD_PRIORITY = 'Ad Priority'
  AD_LANDING_PAGE_ID = 'Landing Page ID'
  AD_CREATIVE_ROTATION_START_TIME = 'Start Date'
  AD_CREATIVE_ROTATION_END_TIME = 'End Date'
  AD_HARDCUTOFF = 'Hard Cutoff'
  AD_TYPE = 'Ad Type'
  AD_URL_SUFFIX = 'URL Suffix'
  CUSTOM_CLICK_THROUGH_URL = 'Custom URL'

  ADVERTISER_ID = 'Advertiser ID'

  ASSET_SIZE = 'Asset Size'

  CAMPAIGN_ID = 'Campaign ID'
  CAMPAIGN_NAME = 'Campaign Name'
  CAMPAIGN_START_DATE = 'Campaign Start Date'
  CAMPAIGN_END_DATE = 'Campaign End Date'
  CAMPAIGN_LANDING_PAGE_NAME = 'Landing Page Name'
  CAMPAIGN_LANDING_PAGE_URL = 'Landing Page URL'
  CAMPAIGN_LANDING_PAGE_ID = 'Landing Page ID'

  CAMPAIGN_CREATIVE_ASSOCIATION_ID = 'Campaign Creative Association ID'
  CREATIVE_ASSET_BUCKET_NAME = 'Creative Asset Bucket Name'
  CREATIVE_ASSET_ID = 'Creative Asset ID'
  CREATIVE_ASSET_NAME = 'Creative Asset Name'
  CREATIVE_ASSET_FILE_NAME = 'Creative Asset File Name'

  CREATIVE_NAME = 'Creative Name'
  CREATIVE_TYPE = 'Creative Type'
  CREATIVE_ID = 'Creative ID'
  CREATIVE_ROTATION = 'Creative Rotation'
  CREATIVE_ROTATION_SEQUENCE = 'Creative Rotation Sequence'
  CREATIVE_ROTATION_WEIGHT = 'Creative Rotation Weight'
  CREATIVE_WIDTH = 'Width (Display Only)'
  CREATIVE_HEIGHT = 'Height (Display Only)'

  CREATIVE_BACKUP_ASSET_ID = 'Backup Asset ID'
  CREATIVE_BACKUP_NAME = 'Creative Backup Name'
  BACKUP_IMAGE_FEATURES = 'Backup Image Features'
  BACKUP_IMAGE_TARGET_WINDOW_OPTION = 'Backup Image Target Window Option'
  BACKUP_IMAGE_CUSTOM_HTML = 'Backup Image Custom HTML'
  BACKUP_IMAGE_CLICK_THROUGH_LANDING_PAGE_ID = ('Backup Image Click Through '
                                                'Landing Page ID')
  BACKUP_IMAGE_CLICK_THROUGH_LANDING_PAGE_NAME = ('Backup Image Click Through '
                                                  'Landing Page Name')
  CLICK_TAGS = 'clickTags'
  CLICK_TAG_NAME = 'Click Tag Name'
  CLICK_TAG_EVENT = 'Click Tag Event'
  CLICK_TAG_LANDING_PAGE_ID = 'Click Tag Landing Page ID'
  CLICK_TAG_LANDING_PAGE_NAME = 'Click Tag Landing Page Name'
  CLICK_TAG_CUSTOM_CLICK_THROUGH_URL = 'Custom Click Through URL'

  PLACEMENT_GROUP_ID = 'Placement Group ID'
  PLACEMENT_GROUP_NAME = 'Placement Group Name'
  PLACEMENT_GROUP_TYPE = 'Placement Group Type'
  PLACEMENT_GROUP_START_DATE = 'Placement Group Start Date'
  PLACEMENT_GROUP_END_DATE = 'Placement Group End Date'
  PLACEMENT_GROUP_PRICING_TYPE = 'Pricing Type'

  PLACEMENT_ID = 'Placement ID'
  PLACEMENT_NAME = 'Placement Name'
  PLACEMENT_START_DATE = 'Placement Start Date'
  PLACEMENT_END_DATE = 'Placement End Date'
  PLACEMENT_TYPE = 'Type'
  PLACEMENT_ARCHIVED = 'Archived'
  PLACEMENT_ACTIVE_VIEW_AND_VERIFICATION = 'Active View and Verification'
  PLACEMENT_AD_BLOCKING = 'Ad Blocking'
  PLACEMENT_PRICING_TESTING_START = 'Pricing Schedule Testing Starts'
  PLACEMENT_PRICING_COST_STRUCTURE = 'Pricing Schedule Cost Structure'
  PLACEMENT_PRICING_SCHEDULE_COST_STRUCTURE = 'Pricing Schedule Cost Structure'
  PLACEMENT_PRICING_SCHEDULE_RATE = 'Pricing Period Rate'
  PLACEMENT_PERIOD_START = 'Pricing Period Start Date'
  PLACEMENT_PERIOD_END = 'Pricing Period End Date'
  PLACEMENT_PERIOD_UNITS = 'Pricing Period Units'
  PLACEMENT_PERIOD_RATE = 'Pricing Period Rate'
  PLACEMENT_SKIPPABLE = 'Skippable'
  PLACEMENT_SKIP_OFFSET_SECONDS = 'Skip Offset Seconds'
  PLACEMENT_SKIP_OFFSET_PERCENTAGE = 'Skip Offset Percentage'
  PLACEMENT_PROGRESS_OFFSET_SECONDS = 'Progress Offset Seconds'
  PLACEMENT_ADDITIONAL_KEY_VALUES = 'Additional Key Values'
  PLACEMENT_PROGRESS_OFFSET_PERCENTAGE = 'Progress Offset Percentage'

  EVENT_TAG_NAME = 'Event Tag Name'
  EVENT_TAG_STATUS = 'Event Tag Status'
  EVENT_TAG_TYPE = 'Event Tag Type'
  EVENT_TAG_URL = 'Event Tag URL'
  EVENT_TAG_ID = 'Event Tag ID'
  EVENT_TAG_ENABLED_BY_DEFAULT = 'Enable By Default'
  EVENT_TAG_PROFILE_NAME = 'Event Tag Profile'
  EVENT_TAG_ENABLED = 'Enabled'

  SITE_ID = 'Site ID'

  TRANSCODE_ID = 'Transcode ID'
  TRANSCODE_MIN_WIDTH = 'Min Width'
  TRANSCODE_MAX_WIDTH = 'Max Width'
  TRANSCODE_MIN_HEIGHT = 'Min Height'
  TRANSCODE_MAX_HEIGHT = 'Max Height'
  TRANSCODE_MIN_BITRATE = 'Min Bitrate (kbps)'
  TRANSCODE_MAX_BITRATE = 'Max Bitrate (kbps)'
  TRANSCODE_FORMAT = 'Format'

  THIRD_PARTY_URL_TYPE = '3P URL Type'
  THIRD_PARTY_URL = '3P URL'

  LP_DAWN_CAMPAIGN_ID = 'CM Campaign ID'
  LP_DAWN_EMAIL = 'Email Address'
  LP_DAWN_LP_ID = 'CM Landing Page ID'
  LP_DAWN_URL = 'Desired Landing Page URL'
  LP_DAWN_DATE_FOR_UPDATE = 'Date for Update (Central Time)'
  LP_DAWN_TIME_FOR_UPDATE = 'Time for Update (Central Time)'
  LP_DAWN_STATUS = 'Status'
  LP_DAWN_UPDATE_TIME = 'Update Time'
  LP_DAWN_MESSAGE = 'Message'

  DYNAMIC_TARGETING_KEY_NAME = 'Key Name'
  DYNAMIC_TARGETING_KEY_OBJECT_TYPE = 'Object Type'
  DYNAMIC_TARGETING_KEY_OBJECT_ID = 'Object ID'


class Feed:
  """Maps Bulkdozer feed items to and from dictionaries."""
  """Maps of internal feed names and Bulkdozer feed tabs.

  Each value is a list because if the first options is not found, it falls back
  to the subsequent one. E.g. lading_page_feed will first try to find a tab
  called "Landing Page", if that isn't found it will use the "Campaign" feed
  isntead, this allows end users to customize the tool behavior by adding or
  removing certain tabs.
  """
  _feed_name_tab_map = {
      'ad_feed': ['Ad'],
      'campaign_feed': ['Campaign'],
      'creative_asset_feed': ['Creative Asset'],
      'creative_feed': ['Creative'],
      'event_tag_feed': ['Event Tag'],
      'landing_page_feed': ['Landing Page', 'Campaign'],
      'placement_group_feed': ['Placement Group'],
      'placement_feed': ['Placement'],
      'third_party_url_feed': ['3P URLs'],
      'ad_creative_assignment_feed': ['Ad Creative Assignment', 'Ad'],
      'ad_placement_assignment_feed': ['Ad Placement Assignment', 'Ad'],
      'event_tag_ad_assignment_feed': ['Event Tag Ad Assignment', 'Ad'],
      'placement_pricing_schedule_feed': ['Placement Pricing Schedule'],
      'creative_asset_association_feed': [
          'Asset Creative Assignment', 'Creative'
      ],
      'creative_campaign_association_feed': [
          'Campaign Creative Assignment', 'Creative'
      ],
      'click_tag_feed': ['Click Tag'],
      'transcode_configs_feed': ['Transcode Configuration'],
      'event_tag_profile_feed': ['Event Tag Profile'],
      'lp_dawn': ['Form Responses 1'],
      'lp_dawn_status': ['Status'],
      'dynamic_targeting_key_feed': ['Dynamic Targeting Keys']
  }

  def __init__(self,
               auth,
               trix_id,
               feed_name,
               parse=True,
               spreadsheet=None,
               timezone=None):
    """Initializes the feed with parameters.

    Args:
      auth: The authentication scheme to use based on the json configuration
        file.
      trix_id: Unique identifier of the Google Sheet that represents the
        Bulkdozer feed.
      feed_name: The name of the feed to initialize.
      spreadsheet: Optional, the spreadsheet object representing the Bulkdozer
        feed spreadsheet, useful to limit calls to the sheets API and allow
        multiple Feed objects to use the same spreadsheet instance. This is used
        to determine which tabs exist in the feed so the correct one can be
        selected for the entity this Feed object represents.
    """
    self.auth = auth
    self.trix_id = trix_id
    self.trix_range = 'A1:AZ'
    self.feed_name = feed_name
    self._parse = parse
    self._timezone = timezone or 'America/New_York'

    # TODO: Make sure we only read the spreadsheet object or the list of tabs
    # once
    if spreadsheet:
      self.spreadsheet = spreadsheet
    else:
      self.spreadsheet = sheets_get(self.auth, self.trix_id)

    self.raw_feed = self._get_feed()
    self.feed = self._feed_to_dict(parse=self._parse)

  def update(self):
    """Updates the related Bulkdozer feed item with the values in this object."""
    new_feed = self._dict_to_feed(parse=self._parse)

    if len(new_feed) > 1:
      sheets_write(self.auth, self.trix_id, self.tab_name, self.trix_range,
                   new_feed)

  def _get_feed(self):
    """Fetches the feed based on initialization parameters.

    Returns:
      List of lists that represents the rows and columns of the feed. If the
      feed isn't found returns a list with an empty list.
    """

    if self.feed_name in self._feed_name_tab_map:
      for tab_name in self._feed_name_tab_map[self.feed_name]:
        for sheet in self.spreadsheet['sheets']:
          if sheet['properties']['title'] == tab_name:
            self.tab_name = tab_name
            return sheets_read(self.auth, self.trix_id, tab_name,
                               self.trix_range)

    return [[]]

  def _convert_date(self, value):
    """Converts dates into a Bulkdozer specific format to be written back to the Feed.

    Args:
      value: String representation of the date.

    Returns:
      Bulkdozer string representation of a date. Returns null if the value
      cannot be parsed into a date.
    """
    try:
      result = parser.parse(value)

      if not result.tzinfo:
        result = pytz.timezone(self._timezone).localize(result)

      if ':' in value:
        return result.strftime('%Y-%m-%dT%H:%M:%S.000%z')
      else:
        return result.strftime('%Y-%m-%d')
    except:
      return None

  def _convert_int(self, value):
    """Converts a value into a integer.

    Args:
      value: String representation of a field from the Bulkdozer feed.

    Returns:
      If possible to convert value into an integer, returns the integer
      representation, otherwise None.
    """
    try:
      return int(value)
    except:
      return None

  def _convert_float(self, value):
    """Conversta a value into a float."""
    try:
      return float(value)
    except:
      return None

  def _parse_value(self, value):
    """Parses a string value into a type specific value infering the correct type based on the data.

    Args:
      value: The value to parse.

    Returns:
      The representation of the value in the correct data type.
    """
    if isinstance(value, str):
      if value.upper() == 'TRUE':
        return True
      elif value.upper() == 'FALSE':
        return False
      else:
        int_value = self._convert_int(value)

        if int_value:
          return int_value
        else:
          float_value = self._convert_float(value)

          if float_value:
            return float_value
          else:
            date_value = self._convert_date(value)

            if date_value:
              return date_value
            else:
              return value
    else:
      return value

  def _dict_to_feed(self, parse=True):
    """Turns a feed into a list of strings to be written back to the feed.

    Args:
      feed: Dictionary list to convert into a list of lists of strings.

    Returns:
     List of list of strings representing the values of the feed.

    """
    if not self.raw_feed:
      return []

    headers = self.raw_feed[0]

    row = 1

    for item in self.feed:
      for key in iter(item.keys()):
        if key in headers:
          column = headers.index(key)
          value = self._parse_value(item[key]) if parse else item[key]
          while column >= len(self.raw_feed[row]):
            self.raw_feed[row].append('')

          self.raw_feed[row][column] = value

      row += 1

    return self.raw_feed

  def _feed_to_dict(self, parse=True):
    """Turns a raw feed from Google Sheets into a list of dictionaries.

    Args:
      raw_feed: List of list of strings representing the feed from Google
        Sheets.

    Returns:
      List of dictionaries with the data from the feed
    """
    if not self.raw_feed:
      return None

    result = []

    header = self.raw_feed[0]

    for line in self.raw_feed[1:]:
      if line and ''.join(line):
        i = 0
        item = {}
        for index, column_header in enumerate(header):
          if index < len(line):
            item[column_header] = self._parse_value(
                line[index].strip()) if parse else line[index].strip()
          else:
            item[column_header] = ''

        result.append(item)
      else:
        break

    return result



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import io

from starthinker.util.project import project
from starthinker.util.dcm import conversions_upload
from starthinker.util.bigquery import query_to_rows
from starthinker.util.sheets import sheets_read
from starthinker.util.csv import csv_to_rows

# Possible CSV headers to ignore
CSV_HEADERS = ['user_id', 'encrypted_user_id']


def conversions_download():
  if project.verbose:
    print('CONVERSION DOWNLOAD')

  # pull from bigquery if specified
  if 'bigquery' in project.task:
    if project.verbose:
      print('READING BIGQUERY')
    rows = query_to_rows(
        project.task['auth'],
        project.id,
        project.task['bigquery']['dataset'],
        'SELECT * FROM %s' % project.task['bigquery']['table'],
        legacy=project.task['bigquery'].get('legacy', True))
    for row in rows:
      yield row

  # pull from sheets if specified
  if 'sheets' in project.task:
    if project.verbose:
      print('READING SHEET')
    rows = sheets_read(project.task['auth'], project.task['sheets']['url'],
                       project.task['sheets']['tab'],
                       project.task['sheets']['range'])
    for row in rows:
      yield row

  # pull from csv if specified
  if 'csv' in project.task:
    if project.verbose:
      print('READING CSV FILE')
    with io.open(project.task['csv']['file']) as f:
      for row in csv_to_rows(f):
        if row[0] not in CSV_HEADERS:
          yield row


@project.from_parameters
def conversion_upload():

  rows = conversions_download()

  if project.verbose:
    print('CONVERSION UPLOAD')

  statuses = conversions_upload(project.task['auth'],
                                project.task['account_id'],
                                project.task['activity_id'],
                                project.task['conversion_type'], rows,
                                project.task['encryptionInfo'])

  has_rows = False
  for status in statuses:
    has_rows = True
    if 'errors' in status:
      if project.verbose:
        print('ERROR:', status['conversion']['ordinal'],
              '\n'.join([e['message'] for e in status['errors']]))
    else:
      if project.verbose:
        print('OK:', status['conversion']['ordinal'])

  if not has_rows:
    if project.verbose:
      print('NO ROWS')


if __name__ == '__main__':
  conversion_upload()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

Activity_Metrics_Schema = [{
    'name': 'Click_Through_Conversions',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Click_Through_Revenue',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'View_Through_Conversions',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'View_Through_Revenue',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Click_Through_Conversions_Cross_Environment',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Click_Through_Revenue_Cross_Environment',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Click_Through_Conversion_Events_Cross_Environment',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Total_Conversions_Cross_Environment',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Total_Revenue_Cross_Environment',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Total_Conversion_Events_Cross_Environment',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'View_Through_Conversions_Cross_Environment',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'View_Through_Revenue_Cross_Environment',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'View_Through_Conversion_Events_Cross_Environment',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_Click_Through_Conversions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_Total_Conversions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_View_Through_Conversions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Actions',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Revenue',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Transactions',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Paid_Search_Action_Conversion_Percentage',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Actions',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Paid_Search_Average_Cost_Per_Action',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Paid_Search_Average_Cost_Per_Transaction',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Paid_Search_Average_Dcm_Transaction_Amount',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Revenue',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Paid_Search_Spend_Per_Transaction_Revenue',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Paid_Search_Transaction_Conversion_Percentage',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Paid_Search_Transaction_Revenue_Per_Spend',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Transactions',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Total_Conversions',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Total_Revenue',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}]



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer ( recommended ):

    From Release: pip install starthinker
    From Open Source: pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory ( if pushing local code changes ):

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

  If any recipe task has "auth" set to "user" add user credentials:

    1. Ensure an RECIPE['setup']['auth']['user'] = [User Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_user", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/deploy_commandline.md#optional-setup-user-credentials

--------------------------------------------------------------

  If any recipe task has "auth" set to "service" add service credentials:

    1. Ensure an RECIPE['setup']['auth']['service'] = [Service Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_service", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md

--------------------------------------------------------------

Segmentology DV360

DV360 funnel analysis using Census data.

  - Wait for <b>BigQuery->UNDEFINED->UNDEFINED->Census_Join</b> to be created.
  - Join the <a hre='https://groups.google.com/d/forum/starthinker-assets' target='_blank'>StarThinker Assets Group</a> to access the following assets
  - Copy <a href='https://datastudio.google.com/c/u/0/reporting/4a908845-fdba-4023-9bb7-68666202bafb/page/K15YB/' target='_blank'>DV360 Segmentology</a>. Leave the Data Source as is, you will change it in the next step.
  - Click Edit Connection, and change to <b>BigQuery->UNDEFINED->(field:recipe_slug}->Census_Join</b>.
  - Or give these intructions to the client.

--------------------------------------------------------------

This StarThinker DAG can be extended with any additional tasks from the following sources:
  - https://google.github.io/starthinker/
  - https://github.com/google/starthinker/tree/master/dags

'''

from starthinker.airflow.factory import DAG_Factory

INPUTS = {
  'auth_read': 'user',  # Credentials used for reading data.
  'recipe_timezone': 'America/Los_Angeles',  # Timezone for report dates.
  'recipe_project': '',  # Project ID hosting dataset.
  'auth_write': 'service',  # Authorization used for writing data.
  'recipe_name': '',  # Name of report, not needed if ID used.
  'recipe_slug': '',  # Name of Google BigQuery dataset to create.
  'partners': [],  # DV360 partner id.
  'advertisers': [],  # Comma delimited list of DV360 advertiser ids.
}

RECIPE = {
  'tasks': [
    {
      'dataset': {
        'description': 'Create a dataset for bigquery tables.',
        'hour': [
          4
        ],
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'dataset': {
          'field': {
            'name': 'recipe_slug',
            'kind': 'string',
            'description': 'Place where tables will be created in BigQuery.'
          }
        }
      }
    },
    {
      'bigquery': {
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing function.'
          }
        },
        'function': 'pearson_significance_test',
        'to': {
          'dataset': {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'order': 4,
              'default': '',
              'description': 'Name of Google BigQuery dataset to create.'
            }
          }
        }
      }
    },
    {
      'dbm': {
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 0,
            'default': 'user',
            'description': 'Credentials used for reading data.'
          }
        },
        'report': {
          'filters': {
            'FILTER_PARTNER': {
              'values': {
                'field': {
                  'name': 'partners',
                  'kind': 'integer_list',
                  'order': 5,
                  'default': [
                  ],
                  'description': 'DV360 partner id.'
                }
              }
            },
            'FILTER_ADVERTISER': {
              'values': {
                'field': {
                  'name': 'advertisers',
                  'kind': 'integer_list',
                  'order': 6,
                  'default': [
                  ],
                  'description': 'Comma delimited list of DV360 advertiser ids.'
                }
              }
            }
          },
          'body': {
            'timezoneCode': {
              'field': {
                'name': 'recipe_timezone',
                'kind': 'timezone',
                'description': 'Timezone for report dates.',
                'default': 'America/Los_Angeles'
              }
            },
            'metadata': {
              'title': {
                'field': {
                  'name': 'recipe_name',
                  'kind': 'string',
                  'order': 3,
                  'prefix': 'Segmentology ',
                  'default': '',
                  'description': 'Name of report, not needed if ID used.'
                }
              },
              'dataRange': 'LAST_30_DAYS',
              'format': 'CSV'
            },
            'params': {
              'type': 'TYPE_CROSS_PARTNER',
              'groupBys': [
                'FILTER_PARTNER',
                'FILTER_PARTNER_NAME',
                'FILTER_ADVERTISER',
                'FILTER_ADVERTISER_NAME',
                'FILTER_MEDIA_PLAN',
                'FILTER_MEDIA_PLAN_NAME',
                'FILTER_ZIP_POSTAL_CODE'
              ],
              'metrics': [
                'METRIC_BILLABLE_IMPRESSIONS',
                'METRIC_CLICKS',
                'METRIC_TOTAL_CONVERSIONS'
              ]
            },
            'schedule': {
              'frequency': 'WEEKLY'
            }
          }
        }
      }
    },
    {
      'dbm': {
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 0,
            'default': 'user',
            'description': 'Credentials used for reading data.'
          }
        },
        'report': {
          'name': {
            'field': {
              'name': 'recipe_name',
              'kind': 'string',
              'order': 3,
              'prefix': 'Segmentology ',
              'default': '',
              'description': 'Name of report, not needed if ID used.'
            }
          }
        },
        'out': {
          'bigquery': {
            'auth': {
              'field': {
                'name': 'auth_write',
                'kind': 'authentication',
                'order': 1,
                'default': 'service',
                'description': 'Authorization used for writing data.'
              }
            },
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'order': 4,
                'default': '',
                'description': 'Name of Google BigQuery dataset to create.'
              }
            },
            'table': 'DV360_KPI',
            'header': True,
            'schema': [
              {
                'name': 'Partner_Id',
                'type': 'INTEGER',
                'mode': 'REQUIRED'
              },
              {
                'name': 'Partner',
                'type': 'STRING',
                'mode': 'REQUIRED'
              },
              {
                'name': 'Advertiser_Id',
                'type': 'INTEGER',
                'mode': 'REQUIRED'
              },
              {
                'name': 'Advertiser',
                'type': 'STRING',
                'mode': 'REQUIRED'
              },
              {
                'name': 'Campaign_Id',
                'type': 'INTEGER',
                'mode': 'REQUIRED'
              },
              {
                'name': 'Campaign',
                'type': 'STRING',
                'mode': 'REQUIRED'
              },
              {
                'name': 'Zip',
                'type': 'STRING',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Impressions',
                'type': 'FLOAT',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Clicks',
                'type': 'INTEGER',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Conversions',
                'type': 'FLOAT',
                'mode': 'NULLABLE'
              }
            ]
          }
        }
      }
    },
    {
      'bigquery': {
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Authorization used for writing data.'
          }
        },
        'from': {
          'query': 'SELECT            Partner_Id,            Partner,            Advertiser_Id,            Advertiser,            Campaign_Id,            Campaign,            Zip,            SAFE_DIVIDE(Impressions, SUM(Impressions) OVER(PARTITION BY Advertiser_Id)) AS Impression_Percent,            SAFE_DIVIDE(Clicks, Impressions) AS Click_Percent,            SAFE_DIVIDE(Conversions, Impressions) AS Conversion_Percent,            Impressions AS Impressions          FROM            `{project}.{dataset}.DV360_KPI`;        ',
          'parameters': {
            'project': {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project ID hosting dataset.'
              }
            },
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be created in BigQuery.'
              }
            }
          },
          'legacy': False
        },
        'to': {
          'dataset': {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'description': 'Place where tables will be written in BigQuery.'
            }
          },
          'view': 'DV360_KPI_Normalized'
        }
      }
    },
    {
      'census': {
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Authorization used for writing data.'
          }
        },
        'normalize': {
          'census_geography': 'zip_codes',
          'census_year': '2018',
          'census_span': '5yr'
        },
        'to': {
          'dataset': {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'order': 4,
              'default': '',
              'description': 'Name of Google BigQuery dataset to create.'
            }
          },
          'type': 'view'
        }
      }
    },
    {
      'census': {
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Authorization used for writing data.'
          }
        },
        'correlate': {
          'join': 'Zip',
          'pass': [
            'Partner_Id',
            'Partner',
            'Advertiser_Id',
            'Advertiser',
            'Campaign_Id',
            'Campaign'
          ],
          'sum': [
            'Impressions'
          ],
          'correlate': [
            'Impression_Percent',
            'Click_Percent',
            'Conversion_Percent'
          ],
          'dataset': {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'order': 4,
              'default': '',
              'description': 'Name of Google BigQuery dataset to create.'
            }
          },
          'table': 'DV360_KPI_Normalized',
          'significance': 80
        },
        'to': {
          'dataset': {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'order': 4,
              'default': '',
              'description': 'Name of Google BigQuery dataset to create.'
            }
          },
          'type': 'view'
        }
      }
    }
  ]
}

DAG_FACTORY = DAG_Factory('segmentology_dv360', RECIPE, INPUTS)
DAG = DAG_FACTORY.generate()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.bigquery import table_create
from starthinker.util.data import get_rows
from starthinker.util.data import put_rows
from starthinker.util.google_api import API_DV360
from starthinker.util.google_api.discovery_to_bigquery import Discovery_To_BigQuery
from starthinker.util.project import project
from starthinker.util.regexp import lookup_id
from starthinker.util.sheets import sheets_clear


def partner_clear():
  table_create(
      project.task['auth_bigquery'],
      project.id,
      project.task['dataset'],
      'DV_Partners',
      Discovery_To_BigQuery('displayvideo',
                            'v1').method_schema('partners.list'),
  )

  sheets_clear(project.task['auth_sheets'], project.task['sheet'], 'Partners', 'B2:Z')


def partner_load():
  # write partners to BQ
  rows = API_DV360(project.task['auth_dv'], iterate=True).partners().list().execute()

  put_rows(
      project.task['auth_bigquery'], {
          'bigquery': {
              'dataset':
                  project.task['dataset'],
              'table':
                  'DV_Partners',
              'schema':
                  Discovery_To_BigQuery('displayvideo',
                                        'v1').method_schema('partners.list'),
              'format':
                  'JSON'
          }
      }, rows)

  # write partners to sheet
  rows = get_rows(
      project.task['auth_bigquery'], {
          'bigquery': {
              'dataset':
                  project.task['dataset'],
              'query':
                  "SELECT CONCAT(displayName, ' - ', partnerId), entityStatus  FROM `%s.DV_Partners`"
                  % project.task['dataset'],
              'legacy':
                  False
          }
      })

  put_rows(project.task['auth_sheets'], {
      'sheets': {
          'sheet': project.task['sheet'],
          'tab': 'Partners',
          'range': 'B2'
      }
  }, rows)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from django.core.management.base import BaseCommand, CommandError
from django.conf import settings

from starthinker_ui.recipe.scripts import Script
from starthinker_ui.recipe.dag import script_to_dag


class Command(BaseCommand):
  help = 'Generate Templates For Airflow'

  def handle(self, *args, **kwargs):
    for script in Script.get_scripts():
      if script.get_open_source():
        print('Writing: %s_dag.py' % script.get_tag())
        with open(
          '%s/dags/%s_dag.py' % (
            settings.UI_ROOT,
            script.get_tag()
          ),
          'w'
        ) as dag_file:
          dag_file.write(
            script_to_dag(
              script.get_tag(),
              script.get_name(),
              script.get_description(),
              script.get_instructions(),
              script.get_tasks_and_setup()
            )
          )



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Handles logging actions back to the Bulkdozer feed Log tab."""

import datetime

from starthinker.util.project import project
from starthinker.util.sheets import sheets_write, sheets_clear


class Timer(object):
  """Timer class responsible for measuring run time for performance profiling and optimization."""

  def __init__(self):
    """Constructor."""
    self._timers = {}

  def start_timer(self, timer_name):
    """Initializes a new timer.

    Args:
      timer_name: name of the timer to initialize, if not unique will reset
        existing timer.
    """
    self._timers[timer_name] = datetime.datetime.now()

  def check_timer(self, timer_name):
    """Checks and prints the elapsed time of a given timer.

    Args:
      timer_name: Name of the timer to check and print, it must have been
        initialized with start_timer.
    """
    if timer_name in self._timers:
      elapsed = datetime.datetime.now() - self._timers[timer_name]

      print('%s: %d.%d' % (timer_name, elapsed.seconds,
                           (elapsed.microseconds -
                            (elapsed.seconds * 60 * 1000000)) / 1000))
    else:
      print('timer %s not defined' % timer_name)


class Logger(object):
  """Logger class responsible for logging data into the Bulkdozer feed's Log tab."""

  def __init__(self, flush_threshold=10):
    """Initializes the logger object.

    This object is a signleton, and therefore when the application starts and
    reads the configuration of which sheet ID to use for the Bulkdozer feed, the
    trix_id and auth fields need to be updated.
    """
    self.trix_id = None
    self.auth = None
    self._row = 1
    self._buffer = []
    self.buffered = True
    self._flush_threshold = flush_threshold

  def clear(self):
    """Clears the log tab in the Bulkdozer feed, useful when a new execution is starting."""
    sheets_clear(self.auth, self.trix_id, 'Log', 'A1:B')
    self._row = 1

  def log(self, message):
    """Logs a message to the Bulkdozer feed's Log tab.

    Args:
      message: The message to log to the feed, it will be appended at the bottom
        of the log, after the last message that was written.
    """
    self._buffer.append(
        [datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S.000%z'), message])

    if not self.buffered or (self._flush_threshold and
                             len(self._buffer) >= self._flush_threshold):
      self.flush()

  def flush(self):
    """Flushes the message buffer writing buffered messages to the sheet."""
    if self._buffer:
      sheets_write(
          self.auth, self.trix_id, 'Log', 'A1', self._buffer, append=True)

      self._row += len(self._buffer)

      self._buffer = []


logger = Logger()
timer = Timer()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from random import choice
from googleapiclient import discovery

from django.db import models
from django.conf import settings
from django.contrib.auth.models import BaseUserManager, AbstractBaseUser

from starthinker.util.auth.wrapper import CredentialsUserWrapper
from starthinker_ui.account.apps import USER_BUCKET


def fix_picture(picture_url):
  return picture_url.replace('/photo.jpg', '/s32-c/photo.jpg')


def token_generate(model_class, model_field, length=8):
  token = None
  while not token:
    token = ''.join([
        choice('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789')
        for i in range(length)
    ])
    try:
      if model_class.objects.filter(**{model_field: token}).exists():
        token = None
    except Exception:
      pass
  return token


def get_profile(credentials):
  service = discovery.build('oauth2', 'v2', credentials=credentials)
  return service.userinfo().get().execute()


class AccountManager(BaseUserManager):

  def create_user(self, credentials=None, password=None, profile=None):
    if profile is None:
      profile = get_profile(credentials)

    account = self.model(
        identifier=profile['id'],
        email=self.normalize_email(profile['email']),
        name=profile['given_name'],
        domain=profile.get('hd', ''),
        picture=fix_picture(profile['picture']),
    )
    account.set_credentials(credentials)
    account.set_password(password)
    account.save(using=self._db)
    return account

  def get_or_create_user(self, credentials=None, password=None):
    profile = get_profile(credentials)

    try:
      account = Account.objects.get(identifier=profile['id'])
      account.email = self.normalize_email(profile['email'])
      account.name = profile['given_name']
      domain = profile.get('hd', ''),
      account.picture = fix_picture(profile['picture'])
      account.set_credentials(credentials)
      account.set_password(password)
      account.save(using=self._db)
    except:
      account = self.create_user(credentials, password, profile)
    return account

  def create_superuser(self, credentials, password):
    account = self.create_user(credentials, password)
    account.is_admin = True
    account.save(using=self._db)
    return account


class Account(AbstractBaseUser):
  identifier = models.CharField(max_length=64, unique=True, db_index=True)

  email = models.EmailField(max_length=255, unique=True)
  name = models.CharField(max_length=255, blank=True, default='')
  domain = models.CharField(max_length=255, blank=True, default='')
  picture = models.CharField(max_length=255, blank=True, default='')

  is_active = models.BooleanField(default=True)
  is_admin = models.BooleanField(default=False)

  birthday = models.DateField(auto_now_add=True)

  objects = AccountManager()

  USERNAME_FIELD = 'email'
  REQUIRED_FIELDS = ['identifier']

  def get_full_name(self):
    return self.email

  def get_short_name(self):
    return self.email

  def __str__(self):
    return self.email

  def __unicode__(self):
    return self.email

  def has_perm(self, perm, obj=None):
    return True

  def has_module_perms(self, app_label):
    return True

  @property
  def is_staff(self):
    return self.is_admin

  def get_profile(self):
    return get_profile(self.get_credentials())

  def get_domain(self):
    try:
      domain = get_profile(self.get_credentials()).get('hd', '')
      if domain != self.domain:
        self.domain = domain
        self.save(update_fields=['domain'])
    except:
      pass
    return self.domain

  def set_credentials(self, credentials):
    # check if refresh token exists before saving credentials ( only given when authenticating not refreshing )
    if self.identifier and credentials.refresh_token:
      buffer = CredentialsUserWrapper()
      buffer.from_credentials(credentials)
      buffer.save(self.get_credentials_path())

  def get_credentials(self):
    return CredentialsUserWrapper(
        self.get_credentials_path()) if self.identifier else None

  def get_credentials_path(self):
    return '%s:ui/%s.json' % (USER_BUCKET, self.identifier)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import re
import pytz
from datetime import date, datetime


RE_LOOKUP = re.compile(r' - (\d+)$')
def lookup_id(lookup):
  result = RE_LOOKUP.search(lookup)
  return result.group(1) if result else None


def date_to_str(value):
  if value is None:
    return None
  else:
    return value.strftime('%Y-%m-%d')


def str_to_date(value):
  if value is None:
    return None
  else:
    return datetime.strptime(value, '%Y-%m-%d').date()


# multiplier is used for milliseconds ( 1000 ) etc...
def datetime_to_epoch(datetime_utc, multiplier=1):
  if datetime_utc is None:
    return None
  else:
    return (datetime_utc - datetime(1970, 1, 1)).total_seconds() * multiplier


# multiplier is used for milliseconds ( 1000 ) etc...
def epoch_to_datetime(epoch_seconds, multiplier=1):
  if epoch_seconds is None:
    return None
  else:
    return datetime.fromtimestamp(int(epoch_seconds) / multiplier, pytz.utc)


RE_YYYYMMDD = re.compile(r'\d{4}[-/_]\d{2}[-/_]\d{2}')


def parse_yyyymmdd(text):
  value = (RE_YYYYMMDD.findall(text) or [None])[0]
  return value


RE__YYYYMMDD = re.compile(r'_?\d{4}[-/]\d{2}[-/]\d{2}')


def strip_yyymmdd(text):
  return RE__YYYYMMDD.sub('', text)


RE_URL = re.compile(r'https?://[^\s\'">]+')


def parse_url(text):
  return RE_URL.findall(text)


RE_DBM_REPORT = re.compile(r'\d{13}_report/')


def parse_dbm_report_id(download_url):
  return (RE_DBM_REPORT.findall(download_url) or
          [''])[0].replace('_report/', '')


RE_ALPHA_NUMERIC = re.compile('([^\s\w]|_)+')


def parse_filename(text):
  return RE_ALPHA_NUMERIC.sub('', text).lower().replace(' ', '_')


#RE_TABLE = re.compile(r'[\[`]([-\w]+)[:\.]([-\w]+)\.([-\w]+)[\]`]')



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.project import project, get_project
from starthinker.util.data import get_rows


def weather_gov_test():
  print('testing weather_gov connector')

  if 'verify' in project.task['weather_gov']:
    rows = get_rows(project.task['auth'],
                    project.task['weather_gov']['verify']['read'])
    station_ids = project.task['weather_gov']['verify']['station_ids']

    for idx, row in enumerate(rows):
      if row[1] != station_ids[idx]:
        raise 'Station weather data not loaded correctly'



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

#https://cloud.google.com/logging/docs/reference/v2/rest/v2/LogEntry

import urllib.request, urllib.error, urllib.parse
import json

from starthinker.config import UI_PROJECT, UI_SERVICE
from starthinker.util.project import project
from starthinker.util.google_api import API_StackDriver

LOG_VERSION = '2'  # must be string

MANAGER_START = 'MANAGER_START'
MANAGER_END = 'MANAGER_END'
MANAGER_TIMEOUT = 'MANAGER_TIMEOUT'
MANAGER_ERROR = 'MANAGER_ERROR'
MANAGER_SCALE = 'MANAGER_SCALE'

JOB_UPDATE = 'JOB_UPDATE'
JOB_START = 'JOB_START'
JOB_END = 'JOB_END'
JOB_ERROR = 'JOB_ERROR'
JOB_CANCEL = 'JOB_CANCEL'
JOB_TIMEOUT = 'JOB_TIMEOUT'


class LogSeverity():
  DEFAULT = 0  # The log entry has no assigned severity level.
  DEBUG = 100  # Debug or trace information.
  INFO = 200  # Routine information, such as ongoing status or performance.
  NOTICE = 300  # Normal but significant events, such as start up, shut down, or a configuration change.
  WARNING = 400  # Warning events might cause problems.
  ERROR = 500  # Error events are likely to cause problems.
  CRITICAL = 600  # Critical events cause more severe problems or outages.
  ALERT = 700  # A person must take an action immediately.
  EMERGENCY = 800  # One or more systems are unusable.


VERBOSE = False


def log_verbose(verbose=True):
  global VERBOSE
  VERBOSE = verbose


INSTANCE_NAME = None


def get_instance_name(default='UNKNOWN'):
  global INSTANCE_NAME
  if INSTANCE_NAME is None:
    try:
      return urllib.request.urlopen(
          urllib.request.Request(
              'http://metadata.google.internal/computeMetadata/v1/instance/name',
              headers={'Metadata-Flavor': 'Google'})).read().decode()
    except:
      INSTANCE_NAME = default
  return INSTANCE_NAME


def log_put(event, severity, job=None, text=None, payload=None):
  """Generic log writer used by helper functions.

  Writes to StackDriver.

  Creates a record that can be read using log_get function. Entire recipe is
  logged, worker data and stdout and stderr are added to the JOSN under worker
  key.
  Only JOB_EXCEPTION and MANAGER_EXCEPTION logs to text in case JSON is corrupt,
  everythng else is JSON.

  Do not call this directly, use helper functions instead:
    - log_manager_start
    - log_manager_timeout
    - log_manager_error
    - log_manager_end
    - log_job_dispatch
    - log_job_receive
    - log_job_start
    - log_job_end
    - log_job_error
    - log_job_timeout

  WARNING: Do not corrupt recipe in log code, it is actively being used by
  workers while being logged.

  Args:
    - event ( string ): One of the JOB_* enums.
    - severity ( enum ): Stackdriver severity level.
    - job ( json ): Recipe workflow to execute.
    - text ( string ): Messaging output form the task. Usual stdout and stderr.
    - payload ( json ): Output from the scaler or any generic json payload.
  """

  if VERBOSE:
    print('LOGGING:', event, severity, text or '')

  body = {
      'entries': [{
          'logName': 'projects/%s/logs/StarThinker' % UI_PROJECT,
          'severity': severity,
          'resource': {
              'type': 'project',
              'labels': {
                  'key': UI_PROJECT
              },
          },
          'labels': {
              'version': LOG_VERSION,
              'layer': event.split('_')[0],
              'event': event,
              'instance': get_instance_name(),
          },
          #"operation": {
          #  "id": string
          #  "producer": string
          #  "first": False,
          #  "last": False,
          #},
          # already in recipe worker logging task and instance, does this have additional value?
          #"sourceLocation": {
          #  "file": string,
          #  "line": string,
          #  "function": string
          #},
      }],
      'partialSuccess': False,
      'dryRun': False
  }

  if text is not None:
    body['entries'][0]['textPayload'] = text
  elif payload is not None:
    body['entries'][0]['jsonPayload'] = payload
  else:
    # Removing tasks from job REMOVES ALL POTENTIAL CREDENTIALS IN CODE
    job_buffer = json.loads(
        json.dumps(job, indent=2, sort_keys=True, default=str))
    if 'tasks' in job_buffer['recipe']:
      del job_buffer['recipe']['tasks']
    if 'auth' in job_buffer['recipe']['setup']:
      del job_buffer['recipe']['setup']['auth']
    body['entries'][0]['jsonPayload'] = job_buffer

  project.initialize(_service=UI_SERVICE, _project=UI_PROJECT)
  try:
    API_StackDriver('service').entries().write(body=body).execute()
  except:
    print('LOG EVENT ERROR')


def log_get(recipe_id=[], timezone='America/Los_Angeles', days=1):
  """Returns last actionable job run for a specific recipe or all recipes.

  Pulls status entries from StackDriver in reverse order.  A single recipe may
  be run multiple times for multiple tasks at different hours, do not
  assume a JOB_END means a recipe is complete.  Only way to ensure a recipe is
  complete
  is to compare all tasks run against all tasks in recipe ( not done by log
  code).

  Args: - recipe_id ( string or list ) - Optional, if provided returns a single
  record for a single job. - timezone ( string ) - The local timezone to cast
  all record times into.

  Returns:
    - ( iterator ) - Each log entry.

  """

  body = {
      'resourceNames': ['projects/%s' % UI_PROJECT,],
      'filter':
          '\
       logName="projects/%s/logs/StarThinker" \
       AND labels.version="%s" \
       AND labels.layer="JOB" \
    ' % (UI_PROJECT, LOG_VERSION),
      'orderBy':
          'timestamp desc',
      'pageSize':
          1000
  }

  if recipe_id:
    if isinstance(recipe_id, str):
      recipe_id = [recipe_id]
    body['filter'] += ' AND ( %s )' % ' OR '.join(
        'operation.id="%s"' % r for r in recipe_id)

  project.initialize(_service=UI_SERVICE, _project=UI_PROJECT)
  for entry in API_StackDriver(
      'service', iterate=True).entries().list(body=body).execute():
    yield entry


def log_manager_start():
  log_put(MANAGER_START, LogSeverity.NOTICE, text='')


def log_manager_end():
  log_put(MANAGER_END, LogSeverity.NOTICE, text='')


def log_manager_timeout():
  log_put(MANAGER_TIMEOUT, LogSeverity.NOTICE, text='')


def log_manager_error(error):
  log_put(MANAGER_ERROR, LogSeverity.ALERT, text=error)


def log_manager_scale(record):
  log_put(MANAGER_SCALE, LogSeverity.NOTICE, payload=record)


def log_job_update(job):
  log_put(JOB_UPDATE, LogSeverity.NOTICE, job)


def log_job_start(job):
  log_put(JOB_START, LogSeverity.NOTICE, job)


def log_job_end(job):
  log_put(JOB_END, LogSeverity.NOTICE, job)


def log_job_error(job):
  log_put(JOB_ERROR, LogSeverity.ERROR, job)


def log_job_cancel(job):
  log_put(JOB_CANCEL, LogSeverity.WARNING, job)


def log_job_timeout(job):
  log_put(JOB_TIMEOUT, LogSeverity.ERROR, job)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import json
import math

from django.shortcuts import render
from django.contrib import messages
from django.db import connection, transaction
from django.template.loader import render_to_string
from django.views.decorators.csrf import csrf_exempt
from django.http import HttpResponse, JsonResponse, HttpResponseRedirect, HttpResponseNotFound
from django.conf import settings

from starthinker_ui.account.decorators import permission_admin
from starthinker_ui.recipe.forms_script import ScriptForm
from starthinker_ui.recipe.models import Recipe, utc_milliseconds
from starthinker_ui.recipe.colab import script_to_colab
from starthinker_ui.recipe.dag import script_to_dag
from starthinker_ui.recipe.log import log_manager_scale
from starthinker_ui.recipe.compute import group_instances_list, group_instances_resize


def recipe_list(request):
  recipes = {
      'running': [],
      'paused': [],
      'finished': [],
      'errors': [],
      'manual': []
  }

  if request.user.is_authenticated:
    for recipe in request.user.recipe_set.all():
      if recipe.manual:
        recipes['manual'].append(recipe)
      elif not recipe.active or recipe.get_log()['status'] == 'NEW':
        recipes['paused'].append(recipe)
      elif recipe.get_log()['status'] == 'FINISHED':
        recipes['finished'].append(recipe)
      elif recipe.get_log()['status'] == 'ERROR':
        recipes['errors'].append(recipe)
      else:
        recipes['running'].append(recipe)

  return render(request, 'recipe/recipe_list.html', {'recipes': recipes})


@permission_admin()
def recipe_edit(request, pk=None, manual=False):
  if pk:
    recipe = request.user.recipe_set.get(pk=pk)
    manual = recipe.manual
  else:
    recipe = None

  if request.method == 'POST':
    form_script = ScriptForm(manual, recipe, request.user, request.POST)
    if form_script.is_valid():
      form_script.save()
      messages.success(request, 'Recipe updated.')
      return HttpResponseRedirect(form_script.instance.link_edit())
    else:
      messages.error(
          request,
          'Recipe Script Errors: %s' % ' '.join(form_script.get_errors()))
  else:
    form_script = ScriptForm(
        manual, recipe, request.user, scripts=request.GET.get('scripts', ''))

  return render(request, 'recipe/recipe_edit.html', {
      'form_script': form_script,
      'manual': manual
  })


@permission_admin()
def recipe_manual(request, pk=None):
  return recipe_edit(request, pk=None, manual=True)


@permission_admin()
def recipe_delete(request, pk=None):
  request.user.recipe_set.filter(pk=pk).delete()
  messages.success(request, 'Recipe deleted.')
  return HttpResponseRedirect('/')


@permission_admin()
def recipe_run(request, pk):
  try:
    recipe = request.user.recipe_set.get(pk=pk)
    if recipe.is_running():
      messages.success(
          request,
          'Recipe dispatched, will run once in progress task completes.')
    else:
      messages.success(request,
                       'Recipe dispatched, give it a few minutes to start.')
    recipe.force()
  except Recipe.DoesNotExist as e:
    messages.error(request, str(e))
  return HttpResponseRedirect('/recipe/edit/%s/' % pk)


@permission_admin()
def recipe_cancel(request, pk):
  try:
    recipe = request.user.recipe_set.get(pk=pk)
    if recipe.is_running():
      messages.success(request,
                       'Recipe cancelled, active task will stop shortly.')
    else:
      messages.success(request, 'Recipe cancelled, no tasks are running.')
    recipe.cancel()
  except Recipe.DoesNotExist as e:
    messages.error(request, str(e))
  return HttpResponseRedirect('/recipe/edit/%s/' % pk)


@permission_admin()
def recipe_status(request, pk):
  try:
    recipe = request.user.recipe_set.get(pk=pk)
    log = recipe.get_log()
    log['report'] = render_to_string('recipe/log.html', {'log': log})
  except Recipe.DoesNotExist:
    log = {}
  return JsonResponse(log)


@csrf_exempt
def recipe_start(request):
  try:
    recipe = Recipe.objects.get(
        reference=request.POST.get('reference', 'invalid'))
    if recipe.is_running():
      response = HttpResponse('RECIPE INTERRUPTED', content_type='text/plain')
    else:
      response = HttpResponse('RECIPE STARTED', content_type='text/plain')
    recipe.force()
  except Recipe.DoesNotExist as e:
    response = HttpResponseNotFound(
        'RECIPE NOT FOUND', content_type='text/plain')
  return response


@csrf_exempt
def recipe_stop(request):
  try:
    recipe = Recipe.objects.get(
        reference=request.POST.get('reference', 'invalid'))
    if recipe.is_running():
      response = HttpResponse('RECIPE INTERRUPTED', content_type='text/plain')
    else:
      response = HttpResponse('RECIPE STOPPED', content_type='text/plain')
    recipe.cancel()
  except Recipe.DoesNotExist as e:
    response = HttpResponseNotFound(
        'RECIPE NOT FOUND', content_type='text/plain')
  return response


@permission_admin()
def recipe_download(request, pk):
  return render(request, 'recipe/download.html', {'recipe': pk})


@permission_admin()
def recipe_json(request, pk):
  try:
    recipe = request.user.recipe_set.get(pk=pk)
    data = recipe.get_json(credentials=False)
    response = HttpResponse(
        json.dumps(data, indent=2), content_type='application/json')
    response[
        'Content-Disposition'] = 'attachment; filename=recipe_%s.json' % recipe.slug(
        )
    return response
  except Exception as e:
    recipe = None
    messages.error(request, str(e))

  return HttpResponseRedirect('/recipe/download/%s/' % pk)


@permission_admin()
def recipe_colab(request, pk):
  try:
    recipe = request.user.recipe_set.get(pk=pk)
    data = script_to_colab(recipe.slug(), '', [],
                           recipe.get_json(credentials=False)['tasks'])
    response = HttpResponse(data, content_type='application/vnd.jupyter')
    response[
        'Content-Disposition'] = 'attachment; filename=colab_%s.ipynb' % recipe.slug(
        )
    return response
  except Exception as e:
    messages.error(request, str(e))
    raise (e)
  return HttpResponseRedirect('/recipe/download/%s/' % pk)


@permission_admin()
def recipe_airflow(request, pk):
  try:
    recipe = request.user.recipe_set.get(pk=pk)
    data = script_to_dag(recipe.slug(), recipe.name, '', [],
                         recipe.get_json(credentials=False)['tasks'])
    response = HttpResponse(data, content_type='application/vnd.jupyter')
    response[
        'Content-Disposition'] = 'attachment; filename=airflow_%s.py' % recipe.slug(
        )
    return response
  except Exception as e:
    messages.error(request, str(e))
    raise (e)
  return HttpResponseRedirect('/recipe/download/%s/' % pk)


def autoscale(request):

  scale = {
      'jobs': 0,
      'workers': {
          'jobs': settings.WORKER_JOBS,
          'max': settings.WORKER_MAX,
          'existing': 0,
          'required': 0
      }
  }

  # get task and worker list
  scale['jobs'] = Recipe.objects.filter(
      active=True, job_utm__lt=utc_milliseconds()).exclude(job_utm=0).count()
  scale['workers']['existing'] = 3 if request == 'TEST' else sum(
      1 for instance in group_instances_list(('PROVISIONING', 'STAGING',
                                              'RUNNING')))
  scale['workers']['required'] = min(
      settings.WORKER_MAX, math.ceil(scale['jobs'] / scale['workers']['jobs']))

  if request != 'TEST' and scale['workers']['required'] > scale['workers'][
      'existing']:
    group_instances_resize(scale['workers']['required'])

  # log the scaling operation
  log_manager_scale(scale)

  return JsonResponse(scale)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from datetime import timedelta, datetime

from starthinker.util.project import project
from starthinker.util.google_api import API_DCM
from starthinker.util.bigquery import table_exists, rows_to_table, query_to_rows
from starthinker.util.dcm import get_profile_for_api
from starthinker.util.data import get_rows

CHANGELOGS_TABLE = 'CM_Change_Logs'
CHANGELOGS_SCHEMA = [
    {
        'name': 'userProfileId',
        'type': 'INTEGER'
    },
    {
        'name': 'accountId',
        'type': 'INTEGER'
    },
    {
        'name': 'subaccountId',
        'type': 'INTEGER'
    },
    {
        'name': 'id',
        'type': 'INTEGER'
    },
    {
        'name': 'transactionId',
        'type': 'INTEGER'
    },
    {
        'name': 'objectType',
        'type': 'STRING'
    },
    {
        'name': 'objectId',
        'type': 'INTEGER'
    },
    {
        'name': 'action',
        'type': 'STRING'
    },
    {
        'name': 'fieldName',
        'type': 'STRING'
    },
    {
        'name': 'changeTime',
        'type': 'TIMESTAMP'
    },
    {
        'name': 'oldValue',
        'type': 'STRING'
    },
    {
        'name': 'newValue',
        'type': 'STRING'
    },
]


def get_changelogs(accounts, start):

  if project.verbose:
    print('CM CHANGE LOGS', accounts)

  for account_id in accounts:

    is_superuser, profile_id = get_profile_for_api(project.task['auth'],
                                                   account_id)
    kwargs = {'profileId': profile_id, 'minChangeTime': start}
    if is_superuser:
      kwargs['accountId'] = account_id

    for changelog in API_DCM(
        'user', iterate=True,
        internal=is_superuser).changeLogs().list(**kwargs).execute():
      yield [
          changelog.get('userProfileId'),
          changelog['accountId'],
          changelog.get('subaccountId'),
          changelog['id'],
          changelog['transactionId'],
          changelog['objectType'],
          changelog['objectId'],
          changelog['action'],
          changelog.get('fieldName'),
          changelog['changeTime'],
          changelog.get('oldValue'),
          changelog.get('newValue'),
      ]


@project.from_parameters
def dcm_log():
  if project.verbose:
    print('DCM LOG')

  accounts = list(get_rows('user', project.task['accounts']))

  # determine start log date
  if table_exists(project.task['out']['auth'], project.task['out']['project'],
                  project.task['out']['dataset'], CHANGELOGS_TABLE):
    start = next(
        query_to_rows(
            project.task['out']['auth'], project.task['out']['project'],
            project.task['out']['dataset'],
            'SELECT FORMAT_TIMESTAMP("%%Y-%%m-%%dT%%H:%%M:%%S-00:00", MAX(changeTime), "UTC") FROM `%s`'
            % CHANGELOGS_TABLE, 1, False))[0]
    disposition = 'WRITE_APPEND'

  else:
    start = (datetime.utcnow() - timedelta(days=int(project.task['days']))
            ).strftime('%Y-%m-%dT%H:%M:%S-00:00')
    disposition = 'WRITE_TRUNCATE'

  # load new logs
  rows = get_changelogs(accounts, start)
  if rows:
    rows_to_table(project.task['out']['auth'], project.task['out']['project'],
                  project.task['out']['dataset'], CHANGELOGS_TABLE, rows,
                  CHANGELOGS_SCHEMA, 0, disposition)


if __name__ == '__main__':
  dcm_log()
